<!--
    This file was copied by hand from
    https://github.com/squidfunk/mkdocs-material/blob/master/material/404.html
    (on Mon 2 Jan 14:48:33 CET 2023), and then modified.
-->
{% block content %}
  <h1>404 - Not found</h1>
{% endblock %}
